from ._linear import LinearRegression
from ._lasso import LassoRegression
from ._dtree import DecisionTreeRegression
from ._rforest import RandomForestRegression
from ._nn import NeuralNetsRegression
from ._nn_wrapper import NeuralNetsWrapper
from ._kneighbors import KNeighborsRegression
from ._svr import SVRegression
